// Copyright 2011, Alan Calvert
// Copyright 2014 - 2024, Will Godfrey

// This file is part of yoshimi, which is free software: you can
// redistribute it and/or modify it under the terms of the GNU General
// Public License as published by the Free Software Foundation, either
// version 2 of the License, or (at your option) any later version.

// yoshimi is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with yoshimi.  If not, see <http://www.gnu.org/licenses/>.


// generated by Fast Light User Interface Designer (fluid) version 1.0405

#ifndef ParametersUI_h
#define ParametersUI_h
#include <FL/Fl.H>
#include <set>
#include "UI/ScaleTrackedWindow.h"
#include "Misc/SynthEngine.h"
#include <FL/Fl_Browser.H>
#include <FL/Fl_Button.H>
#include <FL/Fl_Text_Display.H>

class ParametersUI {
public:
  ScaleTrackedWindow* make_window();
  ScaleTrackedWindow *Recent;
private:
  inline void cb_Recent_i(ScaleTrackedWindow*, void*);
  static void cb_Recent(ScaleTrackedWindow*, void*);
public:
  Fl_Browser *BrowseRecent;
private:
  inline void cb_BrowseRecent_i(Fl_Browser*, void*);
  static void cb_BrowseRecent(Fl_Browser*, void*);
public:
  Fl_Button *Loading;
  Fl_Button *CloseRecent;
private:
  inline void cb_CloseRecent_i(Fl_Button*, void*);
  static void cb_CloseRecent(Fl_Button*, void*);
public:
  Fl_Text_Display *emptytext;
  ParametersUI(SynthEngine *_synth);
  void Show(int _type);
  void Hide();
  ~ParametersUI();
  void paramRtext();
private:
  SynthEngine *synth;
  int H_type;
  int paramDW;
  int paramDH;
public:
  int lastParamW;
};
#endif
